/* automatically generated from ./regs.dat - do not edit */

#ifndef NASM_REGDIS_H
#define NASM_REGDIS_H

#include "regs.h"

extern const enum reg_enum nasm_rd_bndreg  [ 4];
extern const enum reg_enum nasm_rd_creg    [16];
extern const enum reg_enum nasm_rd_dreg    [16];
extern const enum reg_enum nasm_rd_fpureg  [ 8];
extern const enum reg_enum nasm_rd_mmxreg  [ 8];
extern const enum reg_enum nasm_rd_opmaskreg[ 8];
extern const enum reg_enum nasm_rd_reg16   [16];
extern const enum reg_enum nasm_rd_reg32   [16];
extern const enum reg_enum nasm_rd_reg64   [16];
extern const enum reg_enum nasm_rd_reg8    [ 8];
extern const enum reg_enum nasm_rd_reg8_rex[16];
extern const enum reg_enum nasm_rd_sreg    [ 8];
extern const enum reg_enum nasm_rd_treg    [ 8];
extern const enum reg_enum nasm_rd_xmmreg  [32];
extern const enum reg_enum nasm_rd_ymmreg  [32];
extern const enum reg_enum nasm_rd_zmmreg  [32];

#endif /* NASM_REGDIS_H */
