s/@@NASM_MAJOR@@/2/g
s/@@NASM_MINOR@@/11/g
s/@@NASM_SUBMINOR@@/8/g
s/@@NASM_PATCHLEVEL@@/0/g
s/@@NASM_SNAPSHOT@@/0/g
s/@@NASM_VERSION_ID@@/34277376/g
s/@@NASM_VERSION_XID@@/0x020b0800/g
s/@@NASM_VER@@/2.11.08/g
s/@@NASM_MANGLED_VER@@/2.11.08/g
