/* automatically generated from ./regs.dat - do not edit */

#include "tables.h"
#include "nasm.h"

const opflags_t nasm_reg_flags[] = {
    0,
    REG_HIGH,       /* ah    */
    REG_AL,         /* al    */
    REG_AX,         /* ax    */
    REG_HIGH,       /* bh    */
    REG8NA,         /* bl    */
    BNDREG,         /* bnd0  */
    BNDREG,         /* bnd1  */
    BNDREG,         /* bnd2  */
    BNDREG,         /* bnd3  */
    REG16NA,        /* bp    */
    REG8NA,         /* bpl   */
    REG16NA,        /* bx    */
    REG_HIGH,       /* ch    */
    REG_CL,         /* cl    */
    REG_CREG,       /* cr0   */
    REG_CREG,       /* cr1   */
    REG_CREG,       /* cr10  */
    REG_CREG,       /* cr11  */
    REG_CREG,       /* cr12  */
    REG_CREG,       /* cr13  */
    REG_CREG,       /* cr14  */
    REG_CREG,       /* cr15  */
    REG_CREG,       /* cr2   */
    REG_CREG,       /* cr3   */
    REG_CREG,       /* cr4   */
    REG_CREG,       /* cr5   */
    REG_CREG,       /* cr6   */
    REG_CREG,       /* cr7   */
    REG_CREG,       /* cr8   */
    REG_CREG,       /* cr9   */
    REG_CS,         /* cs    */
    REG_CX,         /* cx    */
    REG_HIGH,       /* dh    */
    REG16NA,        /* di    */
    REG8NA,         /* dil   */
    REG_DL,         /* dl    */
    REG_DREG,       /* dr0   */
    REG_DREG,       /* dr1   */
    REG_DREG,       /* dr10  */
    REG_DREG,       /* dr11  */
    REG_DREG,       /* dr12  */
    REG_DREG,       /* dr13  */
    REG_DREG,       /* dr14  */
    REG_DREG,       /* dr15  */
    REG_DREG,       /* dr2   */
    REG_DREG,       /* dr3   */
    REG_DREG,       /* dr4   */
    REG_DREG,       /* dr5   */
    REG_DREG,       /* dr6   */
    REG_DREG,       /* dr7   */
    REG_DREG,       /* dr8   */
    REG_DREG,       /* dr9   */
    REG_DS,         /* ds    */
    REG_DX,         /* dx    */
    REG_EAX,        /* eax   */
    REG32NA,        /* ebp   */
    REG32NA,        /* ebx   */
    REG_ECX,        /* ecx   */
    REG32NA,        /* edi   */
    REG_EDX,        /* edx   */
    REG_ES,         /* es    */
    REG32NA,        /* esi   */
    REG32NA,        /* esp   */
    REG_FS,         /* fs    */
    REG_GS,         /* gs    */
    OPMASK0,        /* k0    */
    OPMASKREG,      /* k1    */
    OPMASKREG,      /* k2    */
    OPMASKREG,      /* k3    */
    OPMASKREG,      /* k4    */
    OPMASKREG,      /* k5    */
    OPMASKREG,      /* k6    */
    OPMASKREG,      /* k7    */
    MMXREG,         /* mm0   */
    MMXREG,         /* mm1   */
    MMXREG,         /* mm2   */
    MMXREG,         /* mm3   */
    MMXREG,         /* mm4   */
    MMXREG,         /* mm5   */
    MMXREG,         /* mm6   */
    MMXREG,         /* mm7   */
    REG64NA,        /* r10   */
    REG8NA,         /* r10b  */
    REG32NA,        /* r10d  */
    REG16NA,        /* r10w  */
    REG64NA,        /* r11   */
    REG8NA,         /* r11b  */
    REG32NA,        /* r11d  */
    REG16NA,        /* r11w  */
    REG64NA,        /* r12   */
    REG8NA,         /* r12b  */
    REG32NA,        /* r12d  */
    REG16NA,        /* r12w  */
    REG64NA,        /* r13   */
    REG8NA,         /* r13b  */
    REG32NA,        /* r13d  */
    REG16NA,        /* r13w  */
    REG64NA,        /* r14   */
    REG8NA,         /* r14b  */
    REG32NA,        /* r14d  */
    REG16NA,        /* r14w  */
    REG64NA,        /* r15   */
    REG8NA,         /* r15b  */
    REG32NA,        /* r15d  */
    REG16NA,        /* r15w  */
    REG64NA,        /* r8    */
    REG8NA,         /* r8b   */
    REG32NA,        /* r8d   */
    REG16NA,        /* r8w   */
    REG64NA,        /* r9    */
    REG8NA,         /* r9b   */
    REG32NA,        /* r9d   */
    REG16NA,        /* r9w   */
    REG_RAX,        /* rax   */
    REG64NA,        /* rbp   */
    REG64NA,        /* rbx   */
    REG_RCX,        /* rcx   */
    REG64NA,        /* rdi   */
    REG_RDX,        /* rdx   */
    REG64NA,        /* rsi   */
    REG64NA,        /* rsp   */
    REG_SEG67,      /* segr6 */
    REG_SEG67,      /* segr7 */
    REG16NA,        /* si    */
    REG8NA,         /* sil   */
    REG16NA,        /* sp    */
    REG8NA,         /* spl   */
    REG_SS,         /* ss    */
    FPU0,           /* st0   */
    FPUREG,         /* st1   */
    FPUREG,         /* st2   */
    FPUREG,         /* st3   */
    FPUREG,         /* st4   */
    FPUREG,         /* st5   */
    FPUREG,         /* st6   */
    FPUREG,         /* st7   */
    REG_TREG,       /* tr0   */
    REG_TREG,       /* tr1   */
    REG_TREG,       /* tr2   */
    REG_TREG,       /* tr3   */
    REG_TREG,       /* tr4   */
    REG_TREG,       /* tr5   */
    REG_TREG,       /* tr6   */
    REG_TREG,       /* tr7   */
    XMM0,           /* xmm0  */
    XMM_L16,        /* xmm1  */
    XMM_L16,        /* xmm10 */
    XMM_L16,        /* xmm11 */
    XMM_L16,        /* xmm12 */
    XMM_L16,        /* xmm13 */
    XMM_L16,        /* xmm14 */
    XMM_L16,        /* xmm15 */
    XMMREG,         /* xmm16 */
    XMMREG,         /* xmm17 */
    XMMREG,         /* xmm18 */
    XMMREG,         /* xmm19 */
    XMM_L16,        /* xmm2  */
    XMMREG,         /* xmm20 */
    XMMREG,         /* xmm21 */
    XMMREG,         /* xmm22 */
    XMMREG,         /* xmm23 */
    XMMREG,         /* xmm24 */
    XMMREG,         /* xmm25 */
    XMMREG,         /* xmm26 */
    XMMREG,         /* xmm27 */
    XMMREG,         /* xmm28 */
    XMMREG,         /* xmm29 */
    XMM_L16,        /* xmm3  */
    XMMREG,         /* xmm30 */
    XMMREG,         /* xmm31 */
    XMM_L16,        /* xmm4  */
    XMM_L16,        /* xmm5  */
    XMM_L16,        /* xmm6  */
    XMM_L16,        /* xmm7  */
    XMM_L16,        /* xmm8  */
    XMM_L16,        /* xmm9  */
    YMM0,           /* ymm0  */
    YMM_L16,        /* ymm1  */
    YMM_L16,        /* ymm10 */
    YMM_L16,        /* ymm11 */
    YMM_L16,        /* ymm12 */
    YMM_L16,        /* ymm13 */
    YMM_L16,        /* ymm14 */
    YMM_L16,        /* ymm15 */
    YMMREG,         /* ymm16 */
    YMMREG,         /* ymm17 */
    YMMREG,         /* ymm18 */
    YMMREG,         /* ymm19 */
    YMM_L16,        /* ymm2  */
    YMMREG,         /* ymm20 */
    YMMREG,         /* ymm21 */
    YMMREG,         /* ymm22 */
    YMMREG,         /* ymm23 */
    YMMREG,         /* ymm24 */
    YMMREG,         /* ymm25 */
    YMMREG,         /* ymm26 */
    YMMREG,         /* ymm27 */
    YMMREG,         /* ymm28 */
    YMMREG,         /* ymm29 */
    YMM_L16,        /* ymm3  */
    YMMREG,         /* ymm30 */
    YMMREG,         /* ymm31 */
    YMM_L16,        /* ymm4  */
    YMM_L16,        /* ymm5  */
    YMM_L16,        /* ymm6  */
    YMM_L16,        /* ymm7  */
    YMM_L16,        /* ymm8  */
    YMM_L16,        /* ymm9  */
    ZMM0,           /* zmm0  */
    ZMM_L16,        /* zmm1  */
    ZMM_L16,        /* zmm10 */
    ZMM_L16,        /* zmm11 */
    ZMM_L16,        /* zmm12 */
    ZMM_L16,        /* zmm13 */
    ZMM_L16,        /* zmm14 */
    ZMM_L16,        /* zmm15 */
    ZMMREG,         /* zmm16 */
    ZMMREG,         /* zmm17 */
    ZMMREG,         /* zmm18 */
    ZMMREG,         /* zmm19 */
    ZMM_L16,        /* zmm2  */
    ZMMREG,         /* zmm20 */
    ZMMREG,         /* zmm21 */
    ZMMREG,         /* zmm22 */
    ZMMREG,         /* zmm23 */
    ZMMREG,         /* zmm24 */
    ZMMREG,         /* zmm25 */
    ZMMREG,         /* zmm26 */
    ZMMREG,         /* zmm27 */
    ZMMREG,         /* zmm28 */
    ZMMREG,         /* zmm29 */
    ZMM_L16,        /* zmm3  */
    ZMMREG,         /* zmm30 */
    ZMMREG,         /* zmm31 */
    ZMM_L16,        /* zmm4  */
    ZMM_L16,        /* zmm5  */
    ZMM_L16,        /* zmm6  */
    ZMM_L16,        /* zmm7  */
    ZMM_L16,        /* zmm8  */
    ZMM_L16,        /* zmm9  */
};
